# -*- coding: utf8 -*-

"""
实现操作系统用户的相关操作
"""

import os
import contextlib
from threading import RLock

_user_sudo_lock = RLock()

@contextlib.contextmanager
def sudo(message="sudo"):
    """临时升级权限到 root .
    """
    # 对于权限这个临界区的访问要串行化
    with _user_sudo_lock as lk:
        # 得到当前进程的 euid
        old_euid = os.geteuid()
        # 提升权限到 root
        os.seteuid(0)
        yield message
        # 恢复到普通权限
        os.seteuid(old_euid)



# -*- encoding: utf-8 -*-

"""dbm-agent 内部数据接口

"""

import logging
from aiohttp import web
from dbma.core.router import routes
from dbma.version import DBM_AGENT_VESION

@routes.view("/apis/dbm-agent")
class DbmAgentView(web.View):
    """
    dbm-agent http 接口实现
    """
    async def get(self):
        """返回 dbm-agent 的版本号信息
        """
        return web.json_response({"name": "dbm-agent", "version": DBM_AGENT_VESION})


# -*- encoding: utf-8 -*-

"""dbm-agent 安装

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import shutil
import logging
import atexit
from pathlib import Path
from dbma.bil.osuser import DBMAUser
from dbma.core.configs import DBM_AGENT_BASE_DIR, DBMAgentConfig
from dbma.bil.net import get_ip_by_card_name


def init(net_card_name:str, dbm_center_url_prefix:str):
    """安装 dbm-agent
    """
    # 配置日志
    logging.basicConfig(level=logging.INFO, format="[%(asctime)s %(levelname)s] - [%(threadName)s] - [%(pathname)s %(lineno)d line]  ~  %(message)s")
    logging.info("start install dbm-agent .")

    # 检查给定的网卡是否存在
    ip = get_ip_by_card_name(net_card_name)
    if ip is None:
        logging.error("not find any ip on {}".format(net_card_name))
        return

    # 创建用户
    logging.info("prepare create user dbma .")
    dbma_user = DBMAUser()
    dbma_user.create()
    logging.info("create user dbma done .")

    # 创建目录和子目录
    logging.info("prepare create directions .")
    if not DBM_AGENT_BASE_DIR.exists():
        DBM_AGENT_BASE_DIR.mkdir()
    
    for subdir in ("etc", "pkgs", "logs", "etc/templates"):
        item = DBM_AGENT_BASE_DIR / subdir
        if not item.exists():
            item.mkdir()
    logging.info("create directions done .")

    # 复制模板文件
    logging.info("prepare copy template files .")
    import dbma
    basedir = Path(dbma.__file__).parent
    src = basedir / "static/cnfs/"
    dest = DBM_AGENT_BASE_DIR / "etc/templates"
    shutil.copytree(src, dest, dirs_exist_ok=True)
    logging.info("copy template files done .")

    # 更新配置并保存到磁盘
    dbm_agent_config = DBMAgentConfig()
    dbm_agent_config.host = ip
    dbm_agent_config.dbmcenter_url_prefix = dbm_center_url_prefix
    dbm_agent_config.sync_to_disk()

    ## 权限调整
    dbma_user.chown(DBM_AGENT_BASE_DIR)

    # 给到启动 dbm-agent 的命令提示
    logging.info("install dbm-agent done . \n dbm-agent start \n to start dbm-agent service . ")



# -*- encoding: utf-8 -*-

"""MySQL 配置文件相关的逻辑

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import re
import os
import json
import logging
import random
from pathlib import Path
from jinja2 import Template
from dataclasses import dataclass, asdict
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.exceptions import MySQLTemplateFileNotExistsException


@dataclass
class MySQLConfig(object):
    """MySQL 配置文件的动态生成
    """
    # basic
    basedir: str = None
    port: str = None
    innodb_buffer_pool_size: str = None

    # region global-config
    version: str = None
    user: str = None
    datadir: str = None
    server_id: int = None
    bind_address: str = "*"
    admin_address: str = "127.0.0.1"
    admin_port: str = None
    mysqlx_port: str = None
    socket: str = None
    mysqlx_socket: str = None
    pid_file: str = None
    character_set_server: str = "utf8mb4"
    collation_server: str = "utf8mb4_0900_ai_ci"
    collation_database: str = "utf8mb4_0900_ai_ci"
    open_files_limit: int = 102400
    max_prepared_stmt_count: int = 4194304
    skip_name_resolve: str = "ON"
    super_read_only: str = "ON"
    read_only: str = "ON"
    log_timestamps: str = "SYSTEM"
    event_scheduler: str = "ON"
    auto_generate_certs: str = "ON"
    activate_all_roles_on_login: str = "OFF"
    end_markers_in_json: str = "OFF"
    tmpdir: str = "/tmp/"
    max_connections: int = 4096
    autocommit: str = "ON"
    sort_buffer_size: int = 262144
    join_buffer_size: int = 262144
    eq_range_index_dive_limit: int = 200
    authentication_policy: str = "*,,"
    show_gipk_in_create_table_and_information_schema: str = "ON"
    sql_generate_invisible_primary_key: str = "ON"

    # endregion global-config

    # region for-tables
    big_tables: str = "ON"
    sql_require_primary_key: str = "ON"
    lower_case_table_names: int = 1
    auto_increment_increment: int = 1
    auto_increment_offset: int = 1
    table_open_cache: int = 10240
    table_definition_cache: int = 2048
    table_open_cache_instances: int = 8
    # endregion for-tables

    # region for-nets
    max_allowed_packet: str = "1G"
    connect_timeout: int = 10
    interactive_timeout: int = 28800
    net_read_timeout: int = 30
    net_retry_count: int = 10
    net_write_timeout: int = 60
    net_buffer_length: int = 32768
    # endregion for-nets

    # region general-logs
    log_output: str = "FILE"
    general_log: str = "OFF"
    general_log_file: str = "general.log"
    # endregion general-logs

    # region slow-logs
    slow_query_log: str = "ON"
    slow_query_log_file: str = "slow.log"
    long_query_time: str = 1
    log_queries_not_using_indexes: str = "OFF"
    log_slow_admin_statements: str = "ON"
    log_slow_replica_statements: str = "ON"
    log_slow_slave_statements: str = "ON"
    # endregion slow-logs

    # region error-logs
    log_error: str = "err.log"
    log_statements_unsafe_for_binlog: str = "ON"
    # endregion error-logs

    # region binlogs
    log_bin: str = None
    binlog_checksum: str = "CRC32"
    log_bin_trust_function_creators: str = "ON"
    binlog_direct_non_transactional_updates: str = "OFF"
    binlog_expire_logs_seconds: int = 2592000
    binlog_error_action: str = "ABORT_SERVER"
    binlog_format: str = "ROW"
    max_binlog_stmt_cache_size: int = 18446744073709547520
    max_binlog_cache_size: int = 18446744073709547520
    max_binlog_size: str = "1GB"
    binlog_order_commits: str = "ON"
    binlog_row_image: str = "FULL"
    binlog_row_metadata: str = "FULL"
    binlog_rows_query_log_events: str = "ON"
    binlog_stmt_cache_size: int = 32768
    log_replica_updates: str = "ON"
    binlog_transaction_compression: str = "ON"
    binlog_transaction_dependency_history_size: int = 25000
    binlog_transaction_dependency_tracking: str = "COMMIT_ORDER"
    sync_binlog: int = 1
    binlog_cache_size: int = 32768
    binlog_group_commit_sync_delay: int = 0
    binlog_group_commit_sync_no_delay_count: int = 0
    # endregion binlogs

    # region gtids
    gtid_mode: str = "ON"
    binlog_gtid_simple_recovery: str = "ON"
    enforce_gtid_consistency: str = "ON"
    gtid_executed_compression_period: int = 0
    # endregion gtids

    # region clone
    plugin_load_add: str = "mysql_clone.so"
    clone_autotune_concurrency: str = "ON"
    clone_buffer_size: int = 4194304
    clone_block_ddl: str = "ON"
    clone_ddl_timeout: int = 300
    clone_enable_compression: str = "OFF"
    clone_max_concurrency: int = 16
    # endregion clone

    # region replication
    rpl_semi_sync_master_enabled: str = "ON"
    rpl_semi_sync_slave_enabled: str = "ON"
    rpl_semi_sync_master_timeout: int = 10000
    rpl_semi_sync_master_wait_point: str = "AFTER_SYNC"
    rpl_semi_sync_master_wait_no_slave: str = "ON"
    rpl_semi_sync_master_wait_for_slave_count: int = 1
    sync_source_info: int = 10000
    skip_replica_start: str = "ON"
    replica_load_tmpdir: str = "/tmp/"
    plugin_load_add: str = "semisync_master.so"
    plugin_load_add: str = "semisync_slave.so"
    relay_log: str = "relay-bin"
    sync_relay_log: int = 10000
    sync_relay_log_info: int = 10000
    slave_preserve_commit_order: str = "ON"
    replica_preserve_commit_order: str = "ON"
    replica_parallel_type: str = "LOGICAL_CLOCK"
    replica_parallel_workers: int = 4
    replica_max_allowed_packet: str = "1G"
    # endregion replication

    # region engines
    default_storage_engine: str = "InnoDB"
    default_tmp_storage_engine: str = "InnoDB"
    internal_tmp_mem_storage_engine: str = "TempTable"
    # endregion engines

    # region innodbs
    innodb_data_home_dir: str = "./"
    innodb_data_file_path: str = "ibdata1:64M:autoextend"
    innodb_page_size: int = 16384
    innodb_default_row_format: str = "DYNAMIC"
    innodb_log_group_home_dir: str = "redo-files"
    innodb_redo_log_capacity: str = "1G"
    innodb_redo_log_encrypt: str = "OFF"
    innodb_online_alter_log_max_size: str = "256M"
    innodb_undo_directory: str = "undo-files"
    innodb_undo_log_encrypt: str = "OFF"
    innodb_undo_log_truncate: str = "ON"
    innodb_max_undo_log_size: str = "1G"
    innodb_rollback_on_timeout: str = "OFF"
    innodb_rollback_segments: int = 128
    innodb_log_checksums: str = "ON"
    innodb_checksum_algorithm: str = "crc32"
    innodb_log_compressed_pages: str = "ON"
    innodb_doublewrite: str = "ON"
    innodb_doublewrite_dir: str = "dblwr-files"
    innodb_doublewrite_files: int = 4
    innodb_doublewrite_batch_size: int = 0
    innodb_commit_concurrency: int = 0
    innodb_read_only: str = "OFF"
    innodb_dedicated_server: str = "OFF"
    innodb_old_blocks_pct: int = 37
    innodb_old_blocks_time: int = 1000
    innodb_random_read_ahead: str = "OFF"
    innodb_read_ahead_threshold: int = 56
    innodb_max_dirty_pages_pct_lwm: int = 10
    innodb_max_dirty_pages_pct: int = 90
    innodb_lru_scan_depth: str = 1024
    innodb_adaptive_flushing: str = "ON"
    innodb_adaptive_flushing_lwm: int = 10
    innodb_flushing_avg_loops: int = 30
    innodb_buffer_pool_dump_pct: int = 25
    innodb_buffer_pool_dump_at_shutdown: str = "ON"
    innodb_buffer_pool_load_at_startup: str = "ON"
    innodb_buffer_pool_filename: str = "ib_buffer_pool"
    innodb_stats_persistent: str = "ON"
    innodb_stats_on_metadata: str = "OFF"
    innodb_stats_method: str = "nulls_equal"
    innodb_stats_auto_recalc: str = "ON"
    innodb_stats_include_delete_marked: str = "OFF"
    innodb_stats_persistent_sample_pages: int = 20
    innodb_stats_transient_sample_pages: int = 8
    innodb_status_output: str = "OFF"
    innodb_status_output_locks: str = "OFF"
    innodb_buffer_pool_dump_now: str = "OFF"
    innodb_buffer_pool_load_abort: str = "OFF"
    innodb_buffer_pool_load_now: str = "OFF"
    innodb_thread_concurrency: int = 0
    innodb_concurrency_tickets: int = 5000
    innodb_thread_sleep_delay: int = 10000
    innodb_adaptive_max_sleep_delay: int = 150000
    innodb_read_io_threads: int = 4
    innodb_write_io_threads: int = 4
    innodb_use_native_aio: str = "ON"
    innodb_flush_sync: str = "ON"
    innodb_spin_wait_delay: int = 6
    innodb_purge_threads: int = 4
    innodb_purge_batch_size: int = 300
    innodb_purge_rseg_truncate_frequency: int = 128
    innodb_deadlock_detect: str = "ON"
    innodb_print_all_deadlocks: str = "OFF"
    innodb_lock_wait_timeout: int = 50
    innodb_table_locks: str = "ON"
    innodb_sync_array_size: int = 1
    innodb_sync_spin_loops: int = 30
    innodb_print_ddl_logs: str = "OFF"
    innodb_replication_delay: int = 0
    innodb_cmp_per_index_enabled: str = "OFF"
    innodb_disable_sort_file_cache: str = "OFF"
    innodb_numa_interleave: str = "OFF"
    innodb_strict_mode: str = "ON"
    innodb_sort_buffer_size: int = 1048576
    innodb_fast_shutdown: int = 0
    innodb_force_load_corrupted: str = "OFF"
    innodb_force_recovery: str = 0
    innodb_temp_tablespaces_dir: str = "innodb-temp-tablespaces"
    innodb_tmpdir: str = "innodb-temps"
    innodb_temp_data_file_path: str = "ibtmp1:12M:autoextend"
    innodb_page_cleaners: int = 4
    innodb_adaptive_hash_index: str = "OFF"
    innodb_adaptive_hash_index_parts: int = 8
    innodb_flush_log_at_timeout: int = 1
    innodb_fsync_threshold: int = 0
    innodb_fill_factor: int = 96
    innodb_file_per_table: str = "ON"
    innodb_autoextend_increment: int = 64
    innodb_open_files: int = 102400
    innodb_buffer_pool_chunk_size: str = "128M"
    innodb_flush_neighbors: int = 0
    innodb_io_capacity: int = 2000
    innodb_io_capacity_max: int = 20000
    innodb_autoinc_lock_mode: int = 2
    innodb_change_buffer_max_size: int = 25
    innodb_flush_method: str = "fsync"
    innodb_flush_log_at_trx_commit: int = 1
    innodb_buffer_pool_instances: int = None
    innodb_log_buffer_size: str = None

    # expire
    innodb_log_files_in_group: int = 8
    innodb_log_file_size: str = "128M"
    # endregion innodbs

    # region ps
    performance_schema: str = "ON"
    # endregion ps

    def calcu_second_attrs(self):
        """根据已有的配置推导出相关的其它配置
        """
        self._calcu_deps_port()
        self._calcu_random_attrs()
        self._calcu_deps_mem()
        self._calcu_deps_basedir()

    def save_to_target_dir(self, target_dir=None):
        """以 json 格式保存到给定的目录
        """
        # 检查目录是否存在
        if not Path(target_dir).exists():
            logging.error(
                "dir '{}' not exists, skip save config to it .".format(target_dir))
            return

        if target_dir is None:
            # 没有给出 target_dir 就保存到 datadir
            json_file = Path(self.datadir) / "mysql-config.json"
        else:
            json_file = Path(target_dir) / "mysql-config.json"

        json_data = asdict(self)
        with open(json_file, 'w') as f:
            f.write(json.dumps(json_data, indent=4))

    def find_mysql_template_file(self):
        """根据版本号加载配置文件模板，如果找不到对应的模板文件就返回 None
        """
        import dbma
        template_file = Path(dbma.__file__).parent / \
            "static/cnfs/mysql-{}.cnf.jinja".format(self.version)
        if template_file.exists():
            return template_file
        logging.warning("template file '{}' not exists ".format(template_file))

        raise MySQLTemplateFileNotExistsException(template_file)

    def render_mysql_template(self, template=None):
        """渲染给定的 template 文件
        """
        if template is None:
            logging.error("template is None .")
            return

        if not template.exists():
            logging.error("template not exists .")
            return

        with open(template, 'r') as f:
            content = f.read()

        t = Template(content)
        return t.render(asdict(self))

    def generate_cnf_config_file(self):
        """ 生成配置文件  /etc/my-{port}-cnf
        """
        # 根据版本号加载配置文件模板
        # 渲染模板
        # 保存渲染后的内容到文件
        try:
            tempate = self.find_mysql_template_file()
        except MySQLTemplateFileNotExistsException as err:
            logging.error(
                "cannot generate cnf config file, becuase {}".format(err))
            raise

        content = self.render_mysql_template(tempate)

        with open(Path("/etc/my-{}.cnf".format(self.port)), 'w') as f:
            f.write(content)

    def generate_init_cnf_config_file(self):
        """生成初始化专用文件, 用于解决 plugin-add 导致的一系列问题

        2023-03-29T11:25:08.320436+08:00 0 [Warning] [MY-013501] [Server] Ignoring --plugin-load[_add] list as the server is running with --initialize(-insecure).
        2023-03-29T11:25:09.326699+08:00 0 [ERROR] [MY-000067] [Server] unknown variable 'rpl_semi_sync_master_enabled=ON'.
        2023-03-29T11:25:09.326717+08:00 0 [ERROR] [MY-013236] [Server] The designated data directory /database/mysql/data/3306/ is unusable. You can remove all files that the server added to it.

        所以我们要在 init 的时候生成一份没有 plugin-add 的版本
        """
        logging.info("starts generate init cnf config file .")
        # 查询 mysql-init 的配置文件模板
        import dbma
        if self.version.startswith("8.0"):
            template_file =  Path(dbma.__file__).parent /  "static/cnfs/mysql-8.0-init-only.jinja"
        else:
            # TODO 支持 5.7.x
            pass

        # 检查模板文件是否存在
        if not template_file.exists():
            raise MySQLTemplateFileNotExistsException(template_file)
        
        # 准备渲染
        with open(template_file) as f:
            content = f.read()
        
        t = Template(content)
        config_content = t.render(asdict(self))

        with open(Path("/tmp/mysql-8.0-init.cnf"), 'w') as f:
            f.write(config_content)
        
        logging.info("starts generate init cnf config file .")

    def generate_systemd_cnf_config(self):
        """
        """
        logging.info("starts generate systemd config file .")
        # 查询 mysql-init 的配置文件模板
        import dbma
        if self.version.startswith("8.0"):
            template_file =  Path(dbma.__file__).parent /  "static/cnfs/mysqld.service.jinja"

        # 检查模板文件是否存在
        if not template_file.exists():
            raise MySQLTemplateFileNotExistsException(template_file)
        
        # 准备渲染
        with open(template_file) as f:
            content = f.read()
        
        t = Template(content)
        config_content = t.render(asdict(self))

        with open(Path("/usr/lib/systemd/system/mysqld-{}.service".format(self.port)), 'w') as f:
            f.write(config_content)
        
        logging.info("ends generate systemd config file .")
        
    def _calcu_deps_port(self):
        """计算依赖于 port 的配置项

        user = 'mysql' + str(port)
        """
        logging.info("starts _calcu_dep_port fun . ")
        # 设置参数
        # 目前 json 模块还不支持序列化 Path 对象，所以这里统一用 str 来表示路径
        self.user = dbm_agent_config.mysql_user_prefix + str(self.port)
        self.datadir = os.path.join(
            dbm_agent_config.mysql_datadir_parent, str(self.port))
        self.admin_port = self.port * 10 + 2
        self.mysqlx_port = self.port * 10
        self.log_bin = os.path.join(
            dbm_agent_config.mysql_binlogdir_parent, str(self.port)+"/binlog")

        # 跟据 datadir 计算 socket 文件的位置
        self.socket = os.path.join(self.datadir, 'mysql.sock')
        self.mysqlx_socket = os.path.join(self.datadir, 'mysqlx.sock')
        self.pid_file = os.path.join(self.datadir, 'mysql.pid')

        logging.info("set user to {} .".format(self.user))
        logging.info("set datadir to {} .".format(self.datadir))
        logging.info("set admin_port to {} .".format(self.admin_port))
        logging.info("set mysqlx_port to {} .".format(self.mysqlx_port))
        logging.info("ends _calcu_dep_port fun . ")

    def _calcu_random_attrs(self):
        """生成随机值
        """
        self.server_id = random.randint(1024, 8192)

    def _calcu_deps_mem(self):
        """计算内存相关的值
        """
        if self.innodb_buffer_pool_size.endswith("M"):
            # M 级别
            self.innodb_buffer_pool_instances = 1
            self.innodb_log_buffer_size = "64M"
        elif self.innodb_buffer_pool_size.endswith("G"):
            # G 级别
            size = re.match("\d*", self.innodb_buffer_pool_size).group(0)
            size = int(size)
            if size <= 2:
                self.innodb_buffer_pool_instances = 1
                self.innodb_log_buffer_size = "64M"
            elif size <= 4:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "128M"
            elif size <= 8:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "256M"
            elif size <= 16:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "512M"
            else:
                self.innodb_buffer_pool_instances = 16
                self.innodb_log_buffer_size = "1G"

    def _calcu_deps_basedir(self):
        """根据 basedir 计算出 MySQL 的 version
        """
        m = re.search(
            r"mysql-(?P<version>\d{1}.\d{1,2}.\d{1,2})-linux", self.basedir)
        self.version = m.group("version")



# -*- encoding: utf-8 -*-

"""dbm-agent http-server

"""

import os
import logging
from aiohttp import web
from logging.handlers import RotatingFileHandler
from dbma.bil.daemon import start_daemon, stop_daemon
from dbma.bil.osuser import get_uid_gid, is_root, DBMAUser
from dbma.core.router import routes
from dbma.core.threads import backends
from dbma.core.configs import DBMAgentConfig
from dbma.core.views import dbmagentview as _
from dbma.components.mysql.views import defaultsview as _


def start():
    """
    """
    # 检查用户
    if not is_root():
        print("please use root use run this program! ")
        exit(1)

    dbma_user = DBMAUser()
    if not dbma_user.is_exists():
        print("please init dbm-agent before use. ")
        exit(2)
    
    # 切换到普通用户
    uid, gid = get_uid_gid(dbma_user.name)
    os.setegid(gid)
    os.seteuid(uid)

    # 以服务运行
    dbm_agent_config = DBMAgentConfig()
    start_daemon(dbm_agent_config.pid_file)
    print("log file '{}' .".format(dbm_agent_config.log_file))

    # 配置日志
    levels = {
        'info': logging.INFO,
        'debug': logging.DEBUG,
        'error': logging.ERROR,
        'warn': logging.WARN
    }
    handler = RotatingFileHandler(filename=dbm_agent_config.log_file,
                                  maxBytes=128 * 1024 * 1024, backupCount=8, encoding="utf8")
    logging.basicConfig(handlers=[handler], level=levels[dbm_agent_config.log_level],
                        format="[%(asctime)s %(levelname)s] - [%(threadName)s] - [%(pathname)s %(lineno)d line]  ~  %(message)s")

    # 服务启动的日志头
    logging.info('-' * 21)
    logging.info("| start dbm-agent . |")
    logging.info('-' * 21)
    logging.info("logging-level {}".format(dbm_agent_config.log_level))

    # 启动后台线程
    logging.info("start backends threads .")
    backends.start_cycle_tasks()

    # 启动 http 服务
    logging.info("going to start dbm-agent http-server bind on 0.0.0.0:8086 .")
    app = web.Application()
    app.add_routes(routes)
    web.run_app(app, host="0.0.0.0", port=8086, access_log=None)


def stop():
    """关闭 dbm-agent 服务
    """
    dbm_agent_config = DBMAgentConfig()
    stop_daemon(dbm_agent_config.pid_file)



# -*- encoding: utf-8 -*-

"""dbm-agent 的配置文件

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import json
import atexit
from pathlib import Path
from dataclasses import dataclass, asdict
from dbma.version import VERSION
from dbma.bil.osuser import DBMAUser
# TODO
# 这里之后要改成 dbma.core.version 文件中的 VERSION 值

DBM_AGENT_BASE_DIR = Path("/usr/local/dbm-agent/")
DBM_AGENT_ETC_DIR = DBM_AGENT_BASE_DIR / "etc"
CONFIG_FILE_PATH = DBM_AGENT_BASE_DIR / "etc/dbm-agent.json"


@dataclass
class DBMAgentConfig(object):
    """
    """
    host: str = "127.0.0.1"
    port: int = 8086
    version: str = VERSION
    dbmcenter_url_prefix: str = "http://127.0.0.1:8080"
    pid_file: str = "/tmp/dbm-agent.pid"
    log_level: str = 'info'
    log_file: str = "/tmp/dbm-agent.log"

    # backends_xxx 后端线程运行周期
    backends_register_time_interval: int = 15

    # 单例模式
    _instance = None

    # MySQL 相关的默认配置
    mysql_datadir_parent: str = "/database/mysql/data/"
    mysql_binlogdir_parent: str = "/database/mysql/binlog/"
    mysql_user_prefix: str = "mysql"
    mysql_default_version: str = "8.0.31"

    def make_register_data(self):
        """
        """
        return {
            'host': self.host,
            'port': self.port,
            'version': self.version
        }

    # 保存配置到磁盘
    def sync_to_disk(self):
        """对象的生命周期结束的时候保存信息到磁盘
        """
        if CONFIG_FILE_PATH.parent.exists():
            config = asdict(self)
            json_str = json.dumps(config, indent=4)
            with open(CONFIG_FILE_PATH, 'w') as c_file:
                c_file.write(json_str)

    # 读取配置文件
    def read_from_disk(self):
        """对象创建的时候从磁盘读取配置文件的内容
        """
        if CONFIG_FILE_PATH.exists():
            json_str = ""
            with open(CONFIG_FILE_PATH, 'r') as c_file:
                json_str = c_file.read()

            try:
                json_data = json.loads(json_str)
                self.__dict__.update(json_data)
            except Exception as err:
                pass

    def __new__(cls, *args, **kw):
        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kw)
        return cls._instance

    # 从配置文件中读取配置项
    def __post_init__(self):
        """从配置文件中读取配置、如果配置文件不存在就使用默认值
        """
        self.read_from_disk()


dbm_agent_config = DBMAgentConfig()


@dataclass
class DBMCenterUrlConfig(object):
    """
    """
    register_agent_url = "{0}/{1}".format(
        dbm_agent_config.dbmcenter_url_prefix, "apis/agents/")

    # 单例模式
    _instance = None

    def __new__(cls, *args, **kw):
        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kw)
        return cls._instance


def _auto_save_to_disk():
    """
    """
    if DBM_AGENT_BASE_DIR.exists():
        dbm_user = DBMAUser()
        dbm_user.chown(DBM_AGENT_BASE_DIR)
        dbm_agent_config.version = VERSION
        dbm_agent_config.sync_to_disk()
        dbm_user.chown(DBM_AGENT_BASE_DIR)


# 进退寻出的时候保存配置对象到磁盘
atexit.register(_auto_save_to_disk)



# -*- encoding: utf-8 -*-

"""MySQL 安装的相关接口
"""

import logging
from aiohttp import web
from pathlib import Path
from dbma.bil.sudos import sudo
from dbma.core.router import routes
from dbma.core.configs import dbm_agent_config
from dbma.core.threads.backends import threads
from dbma.components.mysql.install import install_mysql, uninstall_mysql
from dbma.components.mysql.views.handlers import install_mysql_task_handler


@routes.view("/apis/mysqls/defaults")
class MySQLDefaultsView(web.View):
    """
    MySQL 的全局默认配置
    """
    async def get(self):
        """返回 MySQL 的全局默认配置
        """
        return web.json_response({"mysql_datadir_parent": dbm_agent_config.mysql_datadir_parent,
                                  "default-version": dbm_agent_config.mysql_default_version,
                                  "mysql_binlogdir_parent": dbm_agent_config.mysql_binlogdir_parent})


@routes.view("/apis/mysqls/install")
class MySQLInstallView(web.View):
    """
    MySQL 安装逻辑
    """
    async def post(self):
        logging.info("post-request {}".format(self.request.url))

        data = await self.request.json()

        # 检查参数
        if 'port' not in data:
            logging.warn("port not in post dict")
            return web.json_response({
                'message': "port not in post dict",
            }, status=500)
        port = data['port']

        # 检查 innodb-buffer-pool-size
        if 'ibps' not in data:
            logging.warn("ibps not in post dict")
            return web.json_response({
                'message': "ibps not in post dict",
            }, status=500)
        ibps = data['ibps']

        # 检查包名
        if 'pkg-name' not in data:
            logging.warn("pkg-name not in post dict")
            return web.json_response({
                'message': "pkg-name not in post dict",
            }, status=500)
        pkg_name = data['pkg-name']
        pkg = Path("/usr/local/dbm-agent/pkgs/") / pkg_name

        # 打印一下接收到的参数
        logging.info(
            "port = '{}', ibps = '{}', pkg-name = '{}', pkg = '{}' .".format(port, ibps, pkg_name, pkg))

        # 检查是不是 dbm-center 发过来的请求(如果是它会带上 task_id)
        if 'task-id' in data:
            task_id = data['task-id']
            threads.submit(install_mysql_task_handler, port=port,
                           ibps=ibps, pkg=pkg, task_id=task_id)
            # 准备返回数据
            return web.json_response({
                "message": 'install mysql task is doing .',
            }, status=200)

        try:
            # 临时把权限提升到 root (创建目录、用户 ... 都会用到)
            with sudo("install mysql {}".format(port)):
                install_mysql(port, pkg, innodb_buffer_pool_size=ibps)
            logging.info("install mysql compelete. ")

        except Exception as err:
            logging.info("install mysql got exception {}".format(err))
            return web.json_response({
                "message": str(err),
            }, status=500)
        # 执行到这里说明已经安装完成了
        return web.json_response({
            "message": 'install mysql compelet.',
        }, status=200)


@routes.view("/apis/mysqls/uninstall")
class MySQLUninstallView(web.View):
    """
    MySQL 卸载逻辑
    """
    async def post(self):
        """
        """
        logging.info("post-request {}".format(self.request.url))

        data = await self.request.json()

        # 检查是不是少传了 port 参数
        if 'port' not in data:
            logging.warn("port not in post dict")
            return web.json_response({
                'message': "port not in post dict",
            }, status=500)

        port = data['port']
        try:
            with sudo("install mysql {}".format(port)):
                uninstall_mysql(port)
            logging.info("uninstall mysql complete .")
        except Exception as err:
            logging.error("got error on uninstall mysql {} .".format(port))
            return web.json_response({
                'message': str(err)
            }, status=500)

        # 到这里说明正确的完成了
        return web.json_response({
            'message': "uninstall mysql complete ."
        }, status=200)



# -*- encoding: utf-8 -*-

"""dbm-agent 周期性任务

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""


import time
import atexit
import logging
import requests
from requests.exceptions import ConnectionError
from concurrent.futures import ThreadPoolExecutor
from dbma.core.configs import DBMAgentConfig, DBMCenterUrlConfig

keep_threads_running = True
dbm_center_url_config = DBMCenterUrlConfig()
dbm_agent_config = DBMAgentConfig()
threads = ThreadPoolExecutor(max_workers=2, thread_name_prefix="backends")


def registor_agent_to_center():
    """注册 agent 的信息到 dbm-center
    """
    while keep_threads_running:
        try:
            logging.info("agent ifo  = {} .".format(dbm_agent_config.make_register_data()))
            response = requests.post(dbm_center_url_config.register_agent_url, json=dbm_agent_config.make_register_data())
            data = response.json()
            if data['message'].startswith("UNIQUE constraint failed: agents_agent.host"):
                logging.info("agent has been registed. ")
        except ConnectionError as err:
            logging.info("register agent info to dbm-center '{}' got ConnectionError. maybe dbm-center is done. .".format(dbm_center_url_config.register_agent_url))
        except Exception as err:
            logging.error("registor_agent_to_center fail err-type {}.".format(type(err)))
            logging.exception(err)
        # 默认 15 分钟注册一次
        time.sleep(dbm_agent_config.backends_register_time_interval)


def _stop_threads():
    """关闭后台任务
    """
    global keep_threads_running
    keep_threads_running = False

def start_cycle_tasks():
    """提交所有后台任务到线程池
    """
    threads.submit(registor_agent_to_center)
    atexit.register(_stop_threads)



DBM_AGENT_MAJOR_VERSION = 8
DBM_AGENT_MINOR_VERSION = 31
DBM_AGENT_PATCH_VERSION = 10
GLIB_VERSION = "2.12"

def _get_version_string():
    """
    返回字符串格式的版本号
    """
    return f"{DBM_AGENT_MAJOR_VERSION}.{DBM_AGENT_MINOR_VERSION}.{DBM_AGENT_PATCH_VERSION}"

def _get_default_mysql_pkg():
    """
    返回默认支持到的最新版本的 MySQL 版本号
    """
    return f"mysql-{DBM_AGENT_MAJOR_VERSION}.0.{DBM_AGENT_MINOR_VERSION}-linux-glibc{GLIB_VERSION}-x86_64.tar.xz"

# dbm-agent 版本号
DBM_AGENT_VESION = _get_version_string()
DBM_VERSION_NUMBER = DBM_AGENT_VESION
VERSION = DBM_AGENT_VESION

# 默认支持到的最新版本的 dbm-agent 版本号
DEFAULT_MYSQL_VERSION = _get_default_mysql_pkg()



# -*- coding: utf8 -*-

"""
实现文件系统的相关操作
"""

import os
import tarfile

def is_file_exists(file_path):
    """
    判断文件是否存在
    """
    return os.path.exists(file_path)

def extract_tar_file(tar_file_path, extract_dir):
    """
    解压tar文件

    Parameters:
    ----------
    tar_file_path: str
        等待解压的tar文件路径

    extract_dir: str
        解压到的目标路径

    Return:
    -------
        None
    """
    tar = tarfile.open(tar_file_path)
    tar.extractall(extract_dir)
    tar.close()

def get_tar_file_name(tar_file_path):
    """
    获取tar文件的名称

    Parameters:
    -----------
    tar_file_path: str 
        tar 文件的全路径

    Return:
    -------
        str
    """
    tar = tarfile.open(tar_file_path)
    name = tar.getnames()[0]
    tar.close()
    if os.sep in name:
        name, *_ = name.split(os.sep)
    return name

def link(src, dest):
    """
    创建链接

    Parameters:
    -----------
    src: str
        源路径

    dest: str
        目标路径

    Return:
    -------
        None
    """
    os.symlink(src, dest)

def is_line_in_etc_profile(line):
    """
    /etc/profile 中是否有以 line 开头的行存在

    Parameters:
    -----------
    line: str
        行内容

    Return:
    -------
        bool
    """
    with open("/etc/profile", "r") as f:
        for l in f:
            if l.startswith(line):
                return True
    return False

def append_new_line_to_etc_profile(line):
    """
    在 /etc/profile 中添加一行
    Parameters:
    -----------
    line: str
        要写入的新行
    
    Returns:
    --------
        None
    """
    with open("/etc/profile", "a") as f:
        f.write(line)
        f.write("\n")

join = os.path.join

readlink = os.readlink

listdir = os.listdir

mkdir = os.mkdir


import os
import sys
import stat
import time
import fcntl
import errno
import signal
import argparse
from datetime import datetime
import atexit


__ALL__ = ['start_daemon', 'stop_daemon']


def auto_clean_pid(fileno, pid_file):
    """
    当程序自动退出时清理 pid 文件
    """
    os.close(fileno)
    os.remove(pid_file)


def signal_handler(sig, _):
    """定义信息处理逻辑
    """
    if sig == signal.SIGINT or sig == signal.SIGTERM:
        # signal.SIGINT == 2
        # signal.SIGTERM == 15
        sys.exit(1)


def write_pid_file(pid, pid_file):
    """创建 pid 文件并向其中写入 pid，成功返回 0 异常返回 1
    """
    # 取得 pid 文件的文件描述符
    pid_desc = os.open(pid_file, os.O_CREAT | os.O_RDWR,
                       stat.S_IRUSR | stat.S_IWUSR)
    try:
        fcntl.lockf(pid_desc, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError as err:
        # 排他访问有异常的话，说明已经有一个守护进程在运行了
        print(err)
        return 1
    # 如果能运行到这里说明排他访问是正常的
    # 清空 pid 文件
    os.truncate(pid_desc, 0)
    s_pid = str(pid)
    os.write(pid_desc, s_pid.encode('utf8'))
    atexit.register(auto_clean_pid, pid_desc, pid_file)
    return 0
    # 注意 pid 文件不应该被 close ，因为如果 close 的话其它进程就查询不到是否有进程在用着它了，pid 文件应该是独占的


def start_server(pid_file="/tmp/daemon.pid"):
    """启动服务并把 pid 写入到 pid 文件
    """
    # 注册信号处理函数
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    pid = os.fork()

    # 如果 pid 大于 0 说明这个进程是父进程
    if pid > 0:
        # 等待 5s 这样子进程就应该已经启动了，于是退出父进程
        time.sleep(5)
        sys.exit(0)

    # 如果是父进程它上一个 if 就已经退出了，所以不不会执行到这个，也就是说下面的代码都是子进程的逻辑
    ppid = os.getppid()  # 获取父进程的 id
    pid = os.getpid()   # 获取自己的进程 id

    if write_pid_file(pid, pid_file) != 0:
        # 如果执行到这里说明守护进程已经存在了
        # 那么要退出父进程、退出当前进程
        os.kill(ppid, signal.SIGTERM)
        sys.exit(1)

    # 如果能执行到这里说明当前主机上没有本程序对应的守护进程在运行
    os.setsid()  # 独立成一个新的session,如果不重新设置的会那么session断开之后这个session下的所有进程都会退出。
    signal.signal(signal.SIGHUP, signal.SIG_IGN)
    os.kill(ppid, signal.SIGTERM)  # 主动的kill父进程
    # 后台运行不要接收键盘输入
    sys.stdin.close()


def stop_server(pid_file="/tmp/daemon.pid"):
    """退出守护进程并删除 pid 文件,并且退出当前程序
    """
    try:
        with open(pid_file) as pid_file_ojb:
            s_pid = pid_file_ojb.read()
    except IOError as err:
        # pid 文件都不存在所以守护进程也是不存在的，也就没有必要 kill 它了
        sys.exit(0)
    pid = int(s_pid)
    for i in range(100):
        # kill 100 次，如果都失败了就报 kill 不掉
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError as err:
            if err.errno == errno.ESRCH:
                print("Successful exit")
                break  # 说明上一次已经 kill 成功了
        time.sleep(0.02)
    else:
        sys.exit(1)
    # 删除 pid 文件
    if os.path.isfile(pid_file):
        # 如果守护进程已经主动清理了，那我们就不清理了
        os.remove(pid_file)
    sys.exit(0)


start_daemon = start_server
stop_daemon = stop_server



# -*- encoding: utf-8 -*-

"""MySQL 安装相关的逻辑实现

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import os
import shutil
import tarfile
import logging
from pathlib import Path
from datetime import datetime
from dbma.bil.osuser import MySQLUser
from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.config import MySQLConfig
from dbma.components.mysql.exceptions import MySQLSystemdFileNotExists
from dbma.components.mysql.exceptions import MySQLPkgFileNotExistsException
from dbma.components.mysql.exceptions import InstanceHasBeenInstalledException


default_pkg = Path("/usr/local/dbm-agent/pkgs/mysql-{}-linux-glibc2.12-x86_64.tar.xz".format(
    dbm_agent_config.mysql_default_version))


def pkg_to_basedir(pkg: Path = default_pkg):
    """根据 pkg 的名字计算 basedir 的名字

    Parameter:
    ----------
    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
        Path
    """
    return Path("/usr/local") / (pkg.name.replace('.tar.gz', '').replace('.tar.xz', ''))


def checks_for_install(port: int = 3306, pkg: Path = default_pkg):
    """MySQL 安装前的检查

    Parameters:
    -----------
    port: int
        MySQL 端口

    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
    None

    Exceptions:
    ----------
    MySQLPkgFileNotExistsException

    InstanceHasBeenInstalledException

    """
    # 检查安装包是否存在
    logging.info("check for pkg file exists or not .")
    if not pkg.exists():
        logging.warn("pkg file exists or not .")
        raise MySQLPkgFileNotExistsException(pkg)

    # 检查给定的实例是不是已经安装过了
    logging.info("check {} is installed or not.".format(port))
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / "{}".format(port)
    if datadir.exists():
        logging.warn("instance {} has been installed.".format(port))
        raise InstanceHasBeenInstalledException(str(port))


def check_mysql_systemd_exists(port: int = 3306):
    """检查 mysql 的 systemd 配置文件是否存在

    Paramters:
    ----------
    port: int
        MySQL 端口号

    Returns:
    --------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    logging.info("starts check mysql systemd exists 'mysqld-{}' ".format(port))

    systemd_file = Path("/usr/lib/systemd/system/") / \
        "mysqld-{}.service".format(port)
    if not systemd_file.exists():
        logging.error("systemd config file '{systemd_file}' not exists .")
        raise MySQLSystemdFileNotExists(systemd_file)

    logging.info("ends check mysql systemd exists 'mysqld-{}' ".format(port))


def enable_systemd_for_mysql(port: int = 3306):
    """启用 systemd 服务

    Parameters:
    -----------
    port: int
        MySQL 端口号

    Return:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info("start enable mysql systemd .")
    try:
        check_mysql_systemd_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 enable 操作
        eanble_cmd = "systemctl enable mysqld-{}".format(port)
        logging.info("execute '{}' ".format(eanble_cmd))
        exe_shell_cmd(eanble_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err
    logging.info("ends enable systemd '{}' ".format(port))


def disable_systemd_for_mysql(port: int = 3306):
    """禁用 systemd 服务

    Parameters:
    -----------
    port: int
        MySQL 端口号

    Return:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info("start disable mysql systemd .")
    try:
        check_mysql_systemd_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 enable 操作
        disable_cmd = "systemctl disable mysqld-{}".format(port)
        logging.info("execute '{}' ".format(disable_cmd))
        exe_shell_cmd(disable_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err
    logging.info("ends disable systemd '{}' ".format(port))


def start_mysql(port: int = 3306):
    """启动 MySQL 数据库

    Parameter:
    ----------
    port: int
        MySQL 端口号

    Returns:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info("starts start mysql .")
    try:
        check_mysql_systemd_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 start 操作
        start_cmd = "systemctl start mysqld-{}".format(port)
        logging.info("execute '{}' ".format(start_cmd))
        exe_shell_cmd(start_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err
    logging.info("ends start mysql .")


def stop_mysql(port: int = 3306):
    """关闭 MySQL 数据库

    Parameter:
    ----------
    port: int
        MySQL 端口号

    Returns:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info("starts stop mysql .")
    try:
        check_mysql_systemd_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 stop 操作
        stop_cmd = "systemctl stop mysqld-{}".format(port)
        logging.info("execute '{}' ".format(stop_cmd))
        exe_shell_cmd(stop_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err
    logging.info("ends stop mysql .")


def create_user_and_dirs(port: int = 3306):
    """根据端口号创建 MySQL 用户

    Parameter:
    ----------
    port: int
        MySQL 的端口号

    Return:
    -------
        None
    """
    logging.info("starts create user and dirs port = {} .".format(port))

    # 创建用户
    user = MySQLUser(port)
    user.create()

    # 创建 datadir & binlogdir
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    binlogdir = Path(dbm_agent_config.mysql_binlogdir_parent) / str(port)

    if not datadir.exists():
        logging.info("create datadir '{}' .".format(datadir))
        os.mkdir(datadir)
    else:
        logging.warning("datadir exists, skip cretae it .")

    if not binlogdir.exists():
        logging.info("create binlogdir '{}' .".format(binlogdir))
        os.mkdir(binlogdir)
    else:
        logging.warning("binlogdir exists, skip cretae it .")

    user.chown(datadir)
    user.chown(binlogdir)

    logging.info("ends create user and dirs .".format(port))


def backup_dirs(port: int = 3306, suffix=None):
    """备份数据目录和 binlog 目录，如果 suffix 为 None 就用当前时间值

    Parameters:
    -----------
    port: int
        MySQL 端口号
    
    suffix: str
        备份文件后缀
    
    Returns:
    --------
    None
    """
    # 计算 suffix
    if suffix is None:
        suffix = datetime.now().isoformat().replace(':', '-').replace('.', '-')
    
    # 计算当前的 datadir 和 binlogdir
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    binlogdir = Path(dbm_agent_config.mysql_binlogdir_parent) / str(port)

    datadir_backup_dir = '{}-backup-{}'.format(datadir, suffix)
    binlogdir_backup_dir = '{}-backup-{}'.format(binlogdir, suffix)

    shutil.move(datadir, datadir_backup_dir)
    shutil.move(binlogdir, binlogdir_backup_dir)


def backup_config_file(port: int = 3306, suffix=None):
    """备份配置文件
    """
    # 计算 suffix
    if suffix is None:
        suffix = datetime.now().isoformat().replace(':', '-').replace('.', '-')
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    if not datadir.exists():
        logging.warn("datadir not exists, skip copy config file to datadir .")
        return

    config_file  = "/etc/my-{}.cnf".format(port)
    config_backup_file = datadir / "my-{}.cnf-backup-{}".format(port, suffix)
    shutil.copyfile(config_file, config_backup_file)


def decompression_pkg(pkg: Path = default_pkg):
    """解压安装包到 /usr/local/

    Parameter:
    ---------
    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
        None

    Exception:
    ---------
        MySQLPkgFileNotExistsException
    """
    logging.info("starts decompression pkg .")

    if not pkg.exists():
        logging.error("mysql install package not find '{}' .".format(pkg))

        raise MySQLPkgFileNotExistsException(pkg)

    basedir = pkg_to_basedir(pkg)
    flag_file = basedir / ".dbm-agent-decompression.txt"
    if flag_file.exists():
        logging.info("ends decompression pkg .")
        return

    # 准备解压
    logging.info("open tar file {}".format(pkg))
    with tarfile.open(pkg) as tar_pkg:
        tar_pkg.extractall("/usr/local/")

    # 解压完成之后写入标记文件 basedir/.dbm-agent-decompression.txt
    with open(flag_file, 'w') as f:
        f.write('dbm-agent')

    # 解压完成
    logging.info("ends decompression pkg .")


def create_mysql_config_file(port: int = 3306, basedir: Path = None, innodb_buffer_pool_size: str = "128M"):
    """创建 MySQL 配置文件

    Parameter:
    ----------
    port: int 
        MySQL 端口号、默认值 3306

    basedir: Path
        MySQL basedir

    innodb_buffer_pool_size: int
        缓冲池的大小

    Return:
    -------
        None
    """
    logging.info(
        "starts create mysql config file. basedir = '{}', port = '{}', innodb_buffer_pool_size = '{}' .".format(basedir, port, innodb_buffer_pool_size))

    config = MySQLConfig(basedir=str(basedir), port=port,
                         innodb_buffer_pool_size=innodb_buffer_pool_size)
    config.calcu_second_attrs()
    config.generate_cnf_config_file()
    config.generate_init_cnf_config_file()
    config.generate_systemd_cnf_config()

    logging.info("ends create mysql config file.")


def init_mysql(port: int = 3306, basedir: Path = None):
    """初始化 MySQL

    Parameter:
    ----------
    port: int
        MySQL 端口号

    basedir: Path
        MySQL basedir

    Return:
    -------
        None
    """
    logging.info(
        "starts init mysql port = '{}', basedir = '{}' .".format(port, basedir))

    mysqld = Path(basedir) / "bin/mysqld"
    config = Path("/tmp/mysql-8.0-init.cnf")

    init_cmd = "{} --defaults-file={} --initialize-insecure".format(
        mysqld, config)
    logging.info("init-cmd = '{}' .".format(init_cmd))
    exe_shell_cmd(init_cmd)

    logging.info("ends init mysql.")


def install_mysql(port: int = 3306, pkg: Path = None, innodb_buffer_pool_size: str = "128M"):
    """安装 MySQL 实例

    Parameter:
    ----------
    port: int
        MySQL 端口号

    pkg: Path
        MySQL 安装包全路径

    Returns:
    --------
    None

    Exceptions:
    -----------
    MySQLPkgFileNotExistsException

    InstanceHasBeenInstalledException
    """
    # 安装前的检查
    try:
        checks_for_install(port, pkg)
    except MySQLPkgFileNotExistsException as err:
        logging.error("pkg not exists, skip install . {} .".format(err))
        raise err
    except InstanceHasBeenInstalledException as err:
        logging.error("instance has been installed {} .".format(err))
        raise err
    except Exception as err:
        logging.info("unknown Exception {}".format(err))
        raise err

    # 第一步 创建用户和目录
    create_user_and_dirs(port)

    # 第二步 解压安装包
    decompression_pkg(pkg)

    # 第三步 计算 basedir
    basedir = pkg_to_basedir(pkg)

    # 第四步 创建配置文件
    create_mysql_config_file(port=port, basedir=basedir,
                             innodb_buffer_pool_size=innodb_buffer_pool_size)

    # 第五步 初始化 mysql 实例
    init_mysql(port=port, basedir=basedir)

    # 第六步 启用 mysqld 服务
    enable_systemd_for_mysql(port)

    # 第七步 启动 mysql 实例
    start_mysql(port)

    # export PATH
    # export headers
    # export so


def uninstall_mysql(port: int = 3306):
    """卸载 mysql 数据库

    Exceptions:
    -----------

    MySQLSystemdFileNotExists
    """
    stop_mysql(port)
    disable_systemd_for_mysql(port)
    backup_config_file(port)
    backup_dirs(port)



# -*- coding: utf8 -*-

"""
实现操作系统用户的相关操作
"""

import os
import pwd
import grp
from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.bil.sudos import sudo

def is_root() -> bool:
    """
    检查当前的 euser 是不是 root
    """
    return os.geteuid() == 0


def is_user_exists(user: str) -> bool:
    """检查给定的用户名是否存在

    Parameter
    --------
        user: str 用户名(操作系统层面)

    Return
    ------
        bool
    """

    try:
        pwd.getpwnam(user)
        return True
    except KeyError as err:
        return False
    except TypeError as err:
        return False


def is_group_exists(group: str) -> bool:
    """检查给定的属组是否存在

    Parameter
    ---------
        group: str 属组名

    Return
    ------
        bool
    """
    try:
        grp.getgrnam(group)
        return True
    except KeyError as err:
        return False
    except TypeError as err:
        return False

def get_uid_gid(user_name):
    """
    返回给定用户的 (uid,gid) 组成的元组. 如果给定的用户不存在就返回 (0,0)
    """
    try:
        user = pwd.getpwnam(user_name)
        return user.pw_uid, user.pw_gid
    except Exception as err:
        return 0, 0


class Identify(object):
    """
    """
    _not_implement_message = "please impolement it in sub class ."

    # 标识名(用户名 | 组名)
    name = ''

    def __init__(self, name):
        """
        """
        self.name = name

    def create_shell_str(self) -> str:
        """返回创建对象的 shell 命令行模式

        Return
        ------
            str 返回一条可以执行的 shell 命令
        """
        raise NotImplementedError(self._not_implement_message)

    def drop_shell_str(self) -> str:
        """返回清理对象的 shell 命令行模式

        Return
        ------
            str 返回一条可以执行的 shell 命令
        """
        raise NotImplementedError(self._not_implement_message)

    def is_exists(self):
        """检查给定的标识是否存在

        Return
        ------
            bool  
        """
        raise NotImplementedError(self._not_implement_message)

    def create(self):
        """创建用户|属组

        Return
        ------
            None
        """
        if self.is_exists():
            return
        exe_shell_cmd(self.create_shell_str())

    def drop(self):
        """删除用户|属组

        Return
        ------
            None
        """
        if self.is_exists():
            exe_shell_cmd(self.drop_shell_str())


class BaseGroup(Identify):
    """所有操作系统用户属组的基类
    """
    def __init__(self, name):
        Identify.__init__(self, name)

    def create_shell_str(self) -> str:
        return f"groupadd {self.name}"

    def drop_shell_str(self) -> str:
        return f"groupdel {self.name}"

    def is_exists(self):
        return is_group_exists(self.name)

    def __repr__(self):
        return f"BaseGroup{{name={self.name}}}"

    def __str__(self):
        return f"{self.name}"
    

class BaseUser(Identify):
    """所有用户的基类
    """
    group = None
    home = ''

    def __init__(self, name, home=''):
        Identify.__init__(self,name)

        # 添加定制 home-dir 的支持
        self.home = home

    def create_shell_str(self) -> str:
        if self.home == '':
            return f"useradd {self.name} -g {self.group.name}"
        else:
            return f"useradd {self.name} -g {self.group.name} -d {self.home}"
    
    def drop_shell_str(self) -> str:
        return f"userdel {self.name}"

    def is_exists(self):
        return is_user_exists(self.name)

    def create(self):
        """
        在创建用户的时候如果组不存在，要去创建组
        """
        if (self.group is not None) and (not self.group.is_exists()):
            self.group.create()
        Identify.create(self)

    def chown(self, path, recursive=True):
        """
        """
        if recursive == True:
            cmd = f"chown -R {str(self)} {path}"
        else:
            cmd = f"chown {str(self)} {path}"

        with sudo():
            exe_shell_cmd(cmd)
            
        
    def __str__(self):
        """
        返回 user:group 的形式
        """
        return f"{self.name}:{self.group}"


class DBMAGroup(BaseGroup):
    """
    """
    def __init__(self, name="dbma"):
        BaseGroup.__init__(self,name)


class DBMAUser(BaseUser):
    """
    """
    group = DBMAGroup()

    def __init__(self, name="dbma"):
        BaseUser.__init__(self, name)


class MySQLGroup(BaseGroup):
    def __init__(self,name="mysql"):
        BaseGroup.__init__(self,name)


class MySQLUser(BaseUser):
    """
    """
    # MySQL 端口
    port = 3306
    # 所有的 MySQL 都共用一个 MySQL 组
    group = MySQLGroup()

    def __init__(self,port:int=3306):
        """根据 MySQL 监听的端口创建用户

        Parameter
        ---------
            port: int
        """
        self.name = f"mysql{port}"
        self.port = port
        BaseUser.__init__(self,self.name)
    
    def create(self):
        """创建 MySQL 实例用户(如果属组不存在就先创建属组)
        """
        if self.group.is_exists() == False:
            self.group.create()
        
        BaseUser.create(self)

    def __str__(self):
        return f"{self.name}:{self.group}"


class RootGroup(BaseGroup):
    """
    """    
    def __init__(self, name="root"):
        BaseGroup.__init__(self,name)

    def drop(self):
        """
        root 组是不能删除的、所以这里不做任何实现
        """
        pass


class ZookeeperGroup(BaseGroup):
    """
    """
    def __init__(self, name="zookeeper"):
        BaseGroup.__init__(self,name)


class RootUser(BaseUser):
    group = RootGroup()
    
    def __init__(self):
        BaseGroup.__init__(self,"root")

    def drop(self):
        """
        root 组是不能删除的、所以这里不做任何实现
        """
        pass


class ZookeeperUser(BaseUser):
    group = ZookeeperGroup()
    
    def __init__(self):
        BaseUser.__init__(self,"zookeeper")
   


# -*- coding: utf8 -*-

"""
实现网络相关的能用功能
"""

import psutil
from socket import AddressFamily

def get_ip_by_card_name(name:str):
    """根据给定的网卡名返回对应的 IP 地址，如果网卡名不存在就返回 None

    Paramater:
    ----------
    name: str
        网卡名

    Return
    ------
        str IP 地址(ipv4)
    """
    eths = psutil.net_if_addrs()
    if not name in eths:
        return None

    # 执行到这里说明网卡名存在
    addres = eths.get(name)
    for addr in addres:
        if addr.family == AddressFamily.AF_INET:
            return addr.address
    
    raise ValueError("net card {} not bind with any ip .".format(name))





# -*- coding: utf8 -*-

"""
执行命令行相关的操作都在这里完成
"""

import subprocess
from dbma.bil.sudos import sudo

def exe_shell_cmd(cmd:str):
    """以 root 身份执行命令

    Parameter
    ---------
        cmd:str 要执行的命令

    Return
    ------
        None
    """
    with sudo():
        subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)



import os
from setuptools import setup


def get_version():
    """
    动态获取 dbm-agent 的版本号
    """

    project_dir_name = os.path.dirname(__file__)
    version_file_path = os.path.join(project_dir_name, "dbma/version.py")
    with open(version_file_path) as version_file_obj:
        content = version_file_obj.read()

    g = {}

    exec(content, g, g)
    return g['VERSION']


agent_version = get_version()

setup(name='dbm-agent',
      version=agent_version,
      description='dbm-agent 数据库管理中心客户端程序',
      author="Neeky",
      author_email="neeky@live.com",
      maintainer='Neeky',
      maintainer_email='neeky@live.com',
      scripts=['bin/dbm-agent', 'bin/dbma-cli-single-instance', 'bin/dbma-cli-init',
               'bin/dbm-bt-conn-stack'],
      packages=['dbma', 'dbma/core', 'dbma/core/views', 'dbma/core/agent', 'dbma/core/threads',
                'dbma/bil', 'dbma/components', 'dbma/components/mysql', 'dbma/components/redis',
                'dbma/components/mysql/views'],
      package_data={'dbma': ['static/cnfs/*', 'static/sql-scripts/*']},
      url='https://github.com/Neeky/dbm-agent',
      install_requires=['Jinja2>=2.10.1', 'mysql-connector-python>=8.0.31',
                        'psutil>=5.6.6', 'requests>=2.22.0', 'distro>=1.4.0',
                        'aiohttp>=3.8.1'],
      python_requires='>=3.6.*',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11']
      )



# -*- encoding: utf-8 -*-

"""dbm-agent 收到操作请求(task)之后，如果请求中没有带 task_id ，说明要求同步执行；如果有带 task_id 那么我们要把 task 放到后台线程中执行。

"""

import logging
from pathlib import Path
from typing import Callable
from dbma.bil.sudos import sudo
from dbma.components.mysql.install import install_mysql


def update_task_state_callback(task_id: int = None, state: str = None, message: str = None):
    """调用 dbm-center 的接口更新回调任务的状态
    """
    logging.info("starts update task state callback .")
    # TODO 调用更新接口ends
    logging.info("task_id = {}".format(task_id))
    logging.info("ends update task state callback .")


def install_mysql_task_handler(port: int = 3306, ibps: str = "128M", pkg: Path = None, task_id: int = None):
    """让安装 MySQL 的逻辑放后台执行

    Parameters:
    -----------

    port: int
        MySQL 端口号

    ibps: str
        innodb_buffer_pool_size 的大小， et 128M, 2G, 8G ...

    pkg: str
        MySQL 安装包全路径

    Return:
    -------
    None
    """
    logging.info("starts install mysql task handler .")
    try:
        # 提升后台线程的权限到 root
        with sudo("install mysql task handler"):
            install_mysql(port=port, innodb_buffer_pool_size=ibps, pkg=pkg)
        logging.info("install mysql complete")

        # 是否更新任务信息到 dbm-center
        if not task_id is None:
            logging.warn(
                "install mysql task handler's callback function is None, skip callback")
        else:
            update_task_state_callback(task_id, 200, "install mysql complete")

    except Exception as err:
        logging.error("install mysql task handler got error ")
    logging.info("ends install mysql task handler .")



from dbma.core.exception import DBMAgentException

class MySQLTemplateFileNotExistsException(DBMAgentException):
    """MySQL 配置文件模板不存在
    """
    pass

class MySQLPkgFileNotExistsException(DBMAgentException):
    """MySQL 安装包文件模板不存在
    """
    pass

class InstanceHasBeenInstalledException(DBMAgentException):
    """给定实例已经安装过了
    """
    pass

class MySQLSystemdFileNotExists(DBMAgentException):
    """
    """
    pass